# Copyright (C) 2009 Ferdinand Majerech
# This file is part of IMGCrush
# For conditions of distribution and use, see copyright notice in LICENSE.txt
import sys

from util import *
from help import *
from referencefile import *
from jpg import *
from png import *
from gif import *
from filetester import *
from multiprocessing import *
import re
import globals
import os

class FileAddress:
    """Stores file address for an input file.
    """
    def __init__(self, infname, reffname, outfname):
        """Constructor.
         
        Data members:
        In  : Filename of the input file.
        Ref : Filename of the reference temporary file relative to temp dir.
        Out : Filename of the output file relative to output dir.
        """
        self.In = infname
        self.Ref = reffname
        self.Out = outfname
#TODO: 0.5: in quality mode, if best size is worse than input size, just copy input
#           (input quality is 100% so it's enough for any minquality)
#TODO: 0.5: in size mode, if input is smaller than maxsize and is the smaller
#           than bestfile, just copy input (quality is 100 so it's always the best)
#TODO: 0.5: replace bash commands for things like deleting files,determining filesizes,
#           detecting files etc with python function calls
#TODO: 0.5: refactor common stuff in PNG,JPG,GIF processors to a base class
#TODO: 0.5: Optionally write output files about what settings were used in final images.
#           Use this to detect the unknown GraphicsMagick png grayscale bug.
#TODO: 0.5: Improve API doc (custom stylesheet, index, layout, etc. or something)
#            other than doxygen
#TODO: 0.5: User manual (separate from API doc, detailed info and recommendations
#                        about various options)
class ImageOptimizer:
    """The main class of IMGCrush.

    Handles input and high-level commands to other modules.
    """
    def __init__(self, argv):
        """Constructor.
        
        Initializes the optimizer and its modules and processes input.

        Data members:

        Files and directories:
        SaveDir  : Directory to save output files.
        FileList : Addresses of all input files.

        Resizing:
            By default, resizing preserves aspect ratio, therefore Width and
            Height are maximum bounds of the resized image. If only one
            bound is specified, e.g. Width, the image is resized to fill
            the given bound and the other is resized accordingly to preserve
            aspect ratio. If no bounds are specified or are 0, resizing is 
            turned off. Other members modify resizing behavior. 
        NOTE: Resizing options are deprecated!
        Width           : Maximum width of resized output.
        Height          : Maximum height of resized output.
        Stretch         : If given bounds aren't filled by resized image,
                          stretch the image to fill them.
        
        File format modules:
        UseJPG : Use the JPG module? I.e. consider JPG output for optimization.
        UsePNG : Use the PNG module? I.e. consider PNG output for optimization.
        UseGIF : Use the GIF module? I.e. consider GIF output for optimization.

        Other:
        ProcMult : Multiplier of the number of processes used when processing
                   files. Number of processes actually depends on number of
                   formats used as every format is processed in a separate
                   process. E.g. if ProcMult==2 and we're using JPG and GIF,
                   the actual number of processes is 4
        """
        globals.OUTDIR = "./optimized/"
        globals.TEMPDIR = "./temp/"
        self.FileList = []
        self.Width = "" #DEPRECATED
        self.Height = "" #DEPRECATED
        self.Stretch = False #DEPRECATED
        try:
            self.ProcMult = cpu_count()
            output(str(self.ProcMult) + " cores detected.")
        except NotImplementedError:
            warning("Cannot determine number of CPUs. Setting ProcMult to default value of 2")
            self.ProcMult = 2
        globals.JPG = JPGProcessor()
        self.UseJPG = False
        globals.PNG = PNGProcessor(self.ProcMult)
        self.UsePNG = False
        globals.GIF = GIFProcessor()
        self.UseGIF = False
        globals.TESTER = FileTester()
        self.UseStats = False
        self.RecursiveDirs = False
        self.process_args(argv)
        if not (self.UseJPG or self.UsePNG or self.UseGIF):
            warning("No file format selected. Using default (PNG).")
            self.UsePNG = True
   
    def add_file(self, f, prependdir=None):
        """Adds an input file to the list of files to process.

        Ignores the file if it's invalid.
        
        Arguments:
        f : Filename of the input file. 
        r : Filename of the corresponding temporary reference file
            relative to the temp directory.
        o : Filename of the corresponding output file relative to the optput
            directory.
        """
        output("Adding file " + f)
        for file in self.FileList:
            if file.In == f:
                warning("File " + f + 
                        " specified multiple times in input. Ignoring.")
                return
        o = f
        #remove all dirs from filename
        o = o[o.rfind("/") + 1:]
        if(o.rfind(".") > 0):
            o = o[ :o.rfind(".")]
        if prependdir is not None:
            o = prependdir + o
        #get number of frames: returns 1 on normal images and error on broken ones,
        #only returning more than 1 with multiframe images (which we can't process)
        if run_bash("identify -format '%n' '" + f + "'")== "1":
            self.FileList.append(FileAddress(f, o+"-ref.png",o))
        else :
            output("unreadable or multi frame file: " + f + " :ignoring", 3)
                                                    
    def add_dir(self, dirname):
        """Adds all image files in a directory to the list of files to process.

        Does not add files in subdirectories.

        Arguments:
        dirname : Name of the directory to add files from.
        """
        #always end dir name with / so we can append file name to it
        if dirname[-1] != "/":
            dirname = dirname + "/"
        lscmd = "ls '" + dirname + "'"
        if self.RecursiveDirs:
            lscmd = "ls -R '" + dirname + "'"
        dirls = run_bash(lscmd)
        #if the directory is empty, ignore it
        if(dirls == ""):
            return
        #strip parent dirs to use this as a subdir of temp and save dirs
        absdir = os.path.abspath(dirname) 
        dirstrip = absdir[absdir.rfind("/") :] + "/"
        
        #iterate through files in the directory
        subdir = ""
        for f in dirls.split("\n"):
            #if this is start of a subdirectory:
            if f[0:len(dirname)] == dirname and f[-1] == ":":
                subdir = f[len(dirname):-1]
                subprocess.call("mkdir -p '" + globals.TEMPDIR + dirstrip +
                                subdir + "'", shell=True)
                subprocess.call("mkdir -p '" + globals.OUTDIR + dirstrip + 
                                subdir + "'", shell=True)
                continue
            if f != "":
                self.add_file(dirname + subdir + "/" + f, dirstrip + subdir + "/")
        #make corresponding temp and output directories
        subprocess.call("mkdir -p '" + globals.TEMPDIR + dirstrip+"'", shell = True)
        subprocess.call("mkdir -p '" + globals.OUTDIR + dirstrip + "'", shell = True)
        if not os.path.exists(globals.TEMPDIR + dirstrip) or\
           not os.path.exists(globals.OUTDIR + dirstrip):
               error("Cannot create one of temporary or output directories")
               clean_up()
               sys.exit(-1)
    
    def process_files(self):
        """Generates temp files, selects the best and cleans up.

        This is the main method of imgcrush: 
        it handles generation of files based on input using file format 
        modules, 
        compares them using tester and outputs the best files.
        In the end it deletes the temp directory along with temporary files.

        If multiple processes are used, files are divided into groups, each
        processed by it's own process.
        """
        #number of files already processed
        fcounter = Value("d", 0.0)#0.0
        if self.ProcMult <= 1:
            self.process_file_list(self.FileList, fcounter)
        else:
            #Divide FileList into smaller groups of files
            FileGroups = []
            for g in range(0, self.ProcMult):
                FileGroups.append([])
            for f in range(0, len(self.FileList)):
                FileGroups[f % self.ProcMult].append(self.FileList[f])
            #start a process for every group
            processes = []
            for g in range(0, len(FileGroups)):
                processes.append(Process(target = self.process_file_list, 
                                         args = (FileGroups[g], fcounter, g)))
            for p in processes:
                p.start()
            for p in processes:
                p.join()
        clean_up()
    
    def process_file_list(self, files, fcounter, procnum = 0):
        """Processes given list of files.

        For each file in list, this generates temp files,
        compares them using tester and outputs the best file.
        """
        for file in files:
            reffilename = globals.TEMPDIR + file.Ref
            outfilename = globals.OUTDIR + file.Out
            reffile = ReferenceFile(file.In, reffilename, self.Width, 
                                    self.Height, self.Stretch)
            outfilelist = []
            jqueue = Queue()
            pqueue = Queue()
            gqueue = Queue()
            #If still trouble: check GIF and PNG processes whether or not they terminate correctly
            #initialize processes in self.UseJPG/PNG/GIF
            jpgproc = Process(target = globals.JPG.process,
                              args = (reffile, file.Out, jqueue, procnum))
            pngproc = Process(target = globals.PNG.process,
                              args = (reffile, file.Out, pqueue, procnum))
            gifproc = Process(target = globals.GIF.process,
                              args = (reffile, file.Out, gqueue))
            if self.UseJPG :
                if self.ProcMult > 0:
                    jpgproc.start()
                else:
                    outfilelist.extend(globals.JPG.process(reffile, file.Out))
            if self.UsePNG :
                if self.ProcMult > 0:
                    pngproc.start()
                else:
                    outfilelist.extend(globals.PNG.process(reffile, file.Out))
            if self.UseGIF :
                if self.ProcMult > 0:
                    gifproc.start()
                else:
                    outfilelist.extend(globals.GIF.process(reffile, file.Out))
            if self.ProcMult > 0:
                if self.UseJPG:
                    outfilelist.extend(jqueue.get())
                    jpgproc.join()
                if self.UsePNG:
                    outfilelist.extend(pqueue.get())
                    pngproc.join()
                if self.UseGIF:
                    outfilelist.extend(gqueue.get())
                    gifproc.join()
            globals.TESTER.select_best_file(outfilelist, reffile, outfilename)
            subprocess.call("rm '" + reffile.Name + "'", shell = True)
            fcounter.value += 1.0
            output(str(fcounter.value / len(self.FileList) * 100) + " %")

    def try_tempdir(self, temp):
        """Tries to create given temp directory.

        If temp dir already exists, it appends a number (0) to its name
        and tries to create that, if unsuccesful again, appends incremented
        number (1) and so on. 
        If temp dir can't be created, it returns False, otherwise returns
        its path.
        """
        #basetemp is same as temp except that it never ends with "/" so
        #it can be appended to.
        basetemp = temp
        if(basetemp[-1] == "/"):
            basetemp = basetemp[:-1]
        temp = basetemp
        num = 0
        #if tempdir exists, append a number to it to generate new filename.
        #if that exists, try another number, and so on.
        while os.path.isdir(temp):
            temp = basetemp + str(num)
            num += 1
        temp += "/" 
        try:
            os.makedirs(temp)
            return temp
        except os.error:
            return False

    def init_dirs(self):
        """Initializes temp and output dirs and ends with error if not possible.
        """
        temp = self.try_tempdir(globals.TEMPDIR)
        #if temp dir was succesfully created
        if temp:
            globals.TEMPDIR = temp
        #otherwise try default temp directory
        else:
            warning("Cannot create temp directory: " + globals.TEMPDIR + 
                    " Using default")
            globals.TEMPDIR = "./temp/"
            tempdef = self.try_tempdir(globals.TEMPDIR)
            #if not even default temp directory works, error
            if not tempdef:
                error("Cannot create default temp directory.")
                clean_up()
                sys.exit(-1)
        if globals.OUTDIR[-1] != "/":
            globals.OUTDIR += "/"
        #if output directory does not exist yet, create it
        if not os.path.exists(globals.OUTDIR):
            try:
                os.makedirs(globals.OUTDIR)
            #if we cannot create output directory, try default
            except os.error:
                warning("Cannot create output directory: " + globals.OUTDIR +
                        " Using default.")
                globals.OUTDIR = "./optimized/"
                try:
                    os.makedirs(globals.OUTDIR)
                #if not even default output directory works, error
                except os.error:
                    error("Cannot create default output directory.")
                    clean_up()
                    sys.exit(-1)

    #TODO 0.5: add arguments to boolean options so that stuff can be disabled
    #          as well as enabled
    def process_args(self,argv):                       
        """Processes all command-line input.                

        Input is validated in objects that are affected by it.
        i.e. , input that sets properties of ImageOptimizer is validated here,
        other input is passed to objects that use it and validated there.
        If an argument given to command line option is wrong, default value
        is usually used.                                                 
        If an unrecognizable option is given or an option that requires
        an argument is given without an argument, help is displayed and
        the program is aborted.

        Arguments:
        argv : List of command line options and their arguments.
        """
        opts = []
        args = []
        try:                                
            opts, args = getopt.getopt(argv, globals.OPTS_SHORT, globals.OPTS_LONG)
                                      
        except getopt.GetoptError:
            command_line_help()                          
            sys.exit(2)  
        
        for opt, arg in opts:  
            #Various:
            if opt in ("-h", "--help"):      
                command_line_help()                      
                sys.exit()     
            elif opt in ("-o", "--out-dir"):
                globals.OUTDIR = arg
            elif opt in ("-t", "--temp-dir"):
                globals.TEMPDIR = arg
            elif opt in ("-q", "--quiet"):
                globals.VERBOSITY -= 1
            elif opt in ("-v", "--verbose"):
                globals.VERBOSITY += 1
            elif opt in ("-m", "--prc-mult"):
                self.ProcMult = valid_number(arg, 0, "--prc-mult (-m)", 0,
                                                  16384)
                globals.PNG.set_proc_count(self.ProcMult)
            elif opt in ("-r", "--recur"):
                self.RecursiveDirs = True
            #Formats:
            elif opt in ("-j", "--jpg"): 
                self.UseJPG = True
            elif opt in ("-p", "--png"): 
                self.UsePNG = True      
            elif opt in ("-g", "--gif"): 
                self.UseGIF = True              
            #JPG:
            elif opt in ("-J", "--j-set"):
                globals.JPG.set_all(arg)
            elif opt in ("-U", "--j-qual"):         
                globals.JPG.set_min_quality(arg)
            elif opt in ("-T", "--j-step"): 
                globals.JPG.set_step(arg)
            elif opt in ("-D", "--j-inter"):
                globals.JPG.set_interlace()
            elif opt in ("-O", "--j-optim"):
                globals.JPG.set_optimize()
            elif opt in ("-Z", "--j-s-fact"):
                globals.JPG.set_sfactors(arg)
            elif opt in ("-A","--j-opt-lv"):
                globals.JPG.set_optimizations(arg)
            elif opt in ("-K","--j-gray"):
                globals.JPG.set_grayscale(arg)
            #PNG:
            elif opt in ("-P", "--p-set"):
                globals.PNG.set_all(arg)
            elif opt in ("-L", "--p-lev"): 
                globals.PNG.set_min_level(arg)
            elif opt in ("-H", "--p-filt"): 
                globals.PNG.set_filters(arg)
            elif opt in ("-X", "--p-no-tru"):
                globals.PNG.set_truecolor(False)
            elif opt in ("-Y", "--p-pal"): 
                globals.PNG.set_min_pal_depth(arg)
            elif opt in ("-R", "--p-gray"): 
                globals.PNG.set_min_gray_depth(arg)
            elif opt in ("-E", "--p-inter"): 
                globals.PNG.set_interlace()
            elif opt in ("-B", "--p-opt-lv"):
                globals.PNG.set_optimizations(arg)
            #GIF:
            elif opt in ("-G", "--g-opt"):
                globals.GIF.set_all(arg)
            elif opt in ("-F", "--g-inter"):
                globals.GIF.set_interlace()
            elif opt in ("-I", "--g-depth"):
                globals.GIF.set_min_depth(arg)
            elif opt in ("-C", "--g-opt-lv"):
                globals.GIF.set_optimizations(arg)
            #Optimizaton:
            elif opt in ("-N", "--no-strip"):
                globals.NOSTRIP = True
            elif opt in ("-M", "--mode"): 
                globals.TESTER.set_mode(arg)
            elif opt in ("-Q", "--min-qual"): 
                globals.TESTER.set_min_quality(arg)
            elif opt in ("-S", "--max-size"): 
                globals.TESTER.set_max_size(arg)
            #Resizing (DEPRECATED):
            elif opt in ("-W", "--width"):
                if(arg[-1] == "%"):
                    self.Width = valid_number_str_suffix(arg, "", "%", 
                                                         "--width (-w)", 0, 
                                                         1048576, True)
                else:
                    self.Width = valid_number_str(arg, "", "--width (-w)", 1,
                                                  1048576)
                if self.Width == "0" or self.Width == "0%":
                    self.Width=""
            elif opt in ("-H", "--height"): 
                if(arg[-1]=="%"):
                    self.Height = valid_number_str_suffix(arg, "", "%", 
                                                          "--height (-h)", 0, 
                                                          1048576, True)
                else:
                    self.Height = valid_number_str(arg, "", "--height (-h)", 
                                                   1,1048576)
                if self.Height == "0" or self.Height == "0%":
                    self.Height=""
            elif opt in ("-s", "--stretch"): 
                self.Stretch = True    
              
        self.init_dirs()
        for arg in args:
            if run_bash("[ -d '" + arg + "' ] && echo \"True\"") == "True":
                self.add_dir(arg)
            elif run_bash("[ -f '" + arg + "' ] && echo \"True\"") == "True":
                self.add_file(arg)
            else:
                warning("Input file or directory: " + arg + " does not exist.")
        if len(self.FileList) == 0:
            error("No readable input files")
            command_line_help()
            sys.exit(0)
