# Copyright (C) 2009 Ferdinand Majerech
# This file is part of IMGCrush
# For conditions of distribution and use, see copyright notice in LICENSE.txt

from tempimage import *
from imagesettings import *
from referencefile import *
from multiprocessing import *

#TODO: Optimization: later: Find out which modes (truecolor, gray2, pal8, etc.)
#      work how with which filters.
#      Need better stats module for this as levels are also involved
#TODO: Optimization (lossy): If mode is size and 1bit palette file is too big, it's
#      likely that 2,4,8 palette files will also be too big, so skip them.
#      same for gray. Do this separately for every level/filter combination
#      as other ones can be smaller

class PNGProcData:
    """Data specific to a process using PNGProcessor.

    Contains things like optimization specific data that
    has to be stored separately for each process.
    """
    def __init__(self):
        """Constructor.

        Data members:

        FCount         : Number of generated files already processed by a call
                         to process_png() in this process. Used for % 
                         feedback.
        OptUsePal      : Each entry of this dictionary specifies whether or not
                         to generate indexed files with this bitdepth.
                         None means that this has not been determined yet.
                         This is reset for each call to process_png()
        OptUseGray     : Same as OptUsePal but with grayscale files.
        OptMaxFilesDec : Number of files that were not generated due to
                         optimizations so far. This is used to decrease
                         value returned by get_max_files() to provide
                         precise % feedback.
        DoAdvpng       : Do we need to generate advpng (special case) images?
        """
        self.FCount = 0.0
        self.OptUsePal = { 1: None, 2: None, 4: None, 8: None}
        self.OptUseGray = { 2: None, 4: None, 8: None}
        self.OptMaxFilesDec = 0
        self.DoAdvpng = True

class PNGProcessor:
    """Handles PNG file generation and it's settings.
    """

    def __init__(self, numproc = 1):
        """Constructor.

        Data members:
        MinLevel       : Minimum compression level to use. Can be int 0-9.
        Filters        : PNG filters to use. String of numbers representing 
                         filters, for instance "125" for filters 1, 2 and 5.
                         Available filters:
                         0 : none
                         1 : sub
                         2 : up
                         3 : average
                         4 : Paeth
                         5 : Adaptive when not palleted and level more than 5
                         6 : Adaptive
        MinPalDepth    : Minimum bit depth to use for paletted images. 
                         e.g. bit depth 4 means there is maximum of 16 colors
                         in the palette. Can be 0, 1, 2, 4 or 8. 0 means 
                         paletted images are not generated.
        MinGrayDepth   : Minimum bit depth to use for grayscale images. 
                         e.g. bit depth 4 means there is maximum of 16 levels of
                         gray. Can be 0, 2, 3, 4 or 8. 0 means grayscale images
                         are not generated.
        Interlace      : Generate interlaced images?
        Optimizations  : Level of optimization to use:
                         0 - none
                         1 - lossless optimizations (same results as with none)
                         2 - some optimizations
                         3 - more optimizations
        OptUseGray     : Optimization - dictionary of bools specifying whether
                         or not to generate indexed files with bitdepth equal
                         to dict key. Reset to None with each processes file,
                         which triggers a test that sets it to True or False.
        OptUseGray     : Optimization - dictionary of bools specifying whether
                         or not to generate grayscale files with bitdepth equal
                         to dict key. Reset to None with each processes file,
                         which triggers a test that sets it to True or False.
        OptMaxFilesDec : Number which is substracted from expected number of 
                         generated files in get_max_files() to remove
                         difference caused by optimizations.
        ProcList       : List containing data specific to each process that
                         uses this PNGProcessor.
        """
        self.set_all("light")
        self.TrueColor = True
        self.ProcList=[]
        #set number of processes that will use this PNGProcessor
        self.set_proc_count(numproc)

    def set_all(self, all):
        """Sets all PNG settings according to given mode.
               
        fastest means all optimizations and least generated files
        extreme means no optimizations and all generated files
        heavy means lossless optimizations only - i.e. same result as extreme
        """
        if all == "extreme":
            self.MinLevel = 0
            self.Filters = "0123456"
            self.Interlace = True
            self.MinPalDepth = 1
            self.MinGrayDepth = 2
            self.Optimizations = 0
        elif all == "heavy":
            self.MinLevel = 0
            self.Filters = "0123456"
            self.Interlace = True
            self.MinPalDepth = 1
            self.MinGrayDepth = 2
            self.Optimizations = 1
        elif all == "medium":
            self.MinLevel = 0
            self.Filters = "0146"
            self.Interlace = False
            self.MinPalDepth = 2
            self.MinGrayDepth = 4
            self.Optimizations = 2
        elif all == "light":
            self.MinLevel = 3
            self.Filters = "046"
            self.Interlace = False
            self.MinPalDepth = 4
            self.MinGrayDepth = 4
            self.Optimizations = 3
        elif all == "fastest":
            self.MinLevel = 6
            self.Filters = "06"
            self.Interlace = False
            self.MinPalDepth = 4
            self.MinGrayDepth = 8
            self.Optimizations = 4
        else:
            warning("wrong value passed by --p-setl : using default")

    def set_proc_count(self, numproc):
        """Set the number of processes that will use PNGProcessor.

        Used to allocate process specific data.
        """
        if numproc == 0:
            numproc = 1
        for p in range (0, numproc):
            self.ProcList.append(PNGProcData())

    def set_optimizations(self, opt):
        """Set optimization level to use.
        """
        try:
            self.Optimizations = clamp(int(opt),0,5)
        except ValueException:
            warning("wrong value passed by --p-opt-lv : using default")
    
    def set_min_level(self, l):
        """Sets minimum compression level to use.
        """
        self.MinLevel = valid_number(l, 7, "--p-lev",0,9)
  
    def set_filters(self, filters):
        """Sets PNG filters to use.
        """
        self.Filters = ""
        for f in filters:
            if f in ("0", "1", "2", "3", "4", "5", "6")\
                 and (self.Filters.find(f) == -1):
                self.Filters = self.Filters + f
        if(self.Filters == ""):
            warning("wrong value passed by --p-filt : using default")
            self.Filters = "15"

    def set_min_pal_depth(self, d):
        """Sets minimum bit depth to use with paletted images.
        """
        try:
            if(int(d) in (0, 1, 2, 4, 8)):
                self.MinPalDepth = int(d)
            elif(int(d)==0):
                pass
            else:
                warning("wrong value passed by --p-pal : using default")
        except ValueException:
            warning("wrong value passed by --p-pal : using default")
    
    def set_min_gray_depth(self, d):
        """Sets the minimum bit depth to use with grayscale images.
        """
        try:
            if(int(d) in (0, 2, 4, 8)):
                self.MinGrayDepth = int(d)
            else:
                warning("wrong value passed by --p-gray : using default")
        except ValueException:
            warning("wrong value passed by --p-gray : using default")
    
    def set_interlace(self):
        """Turns generation of interlaced images on.
        """
        self.Interlace = True

    def set_truecolor(self, t):
        """Turns generation of truecolor images on or off.
        """
        self.TrueColor = t

    def get_level_vals(self):
        """Parses stored PNG level settings, returns list of used level values.
        """
        out = range(self.MinLevel, 10)
        #OPTIMIZATION: Do not use PNG level 1 as it never is the best result
        #and is usually the worst result. This might need more statistics
        #to confirm, though.
        #On optimization level 2, do not use PNG level 2 as in never seems to
        #be the best (or worst) result. again, more statistics needed.
        if self.Optimizations >= 1:
            if 1 in out:
                out.remove(1)
        if self.Optimizations >= 2:
            if 2 in out:
                out.remove(2)
        return out

    def get_filter_vals(self):
        """Parses stored filter settings, returns list of used filter values.
        """
        return list(self.Filters)

    def get_mode_vals(self):
        """Parses stored color storage mode data, returns possible values.
        """
        out = ["truecolor"]
        if self.BitMap:
            out.append("bitmap")
        if self.MinPalDepth:
            out.append("palette")
        if self.MinGrayDepth:
            out.append("gray")
        return out
    
    def get_bitdepth_vals(self):
        """Parses stored bitdepth data, returns list of used bitdepth values.
        """
        #truecolor is always used and is 8*3 bpp
        out = [24]
        if self.BitMap or self.MinPalDepth == 1:
            out.append(1)
        if (self.MinPalDepth <= 2) or (self.MinGrayDepth <= 2):
            out.append(2)
        if (self.MinPalDepth <= 4) or (self.MinGrayDepth <= 4):
            out.append(4)
        if (self.MinPalDepth <= 8) or (self.MinGrayDepth <= 8):
            out.append(8)
        return out

    def get_interlace_vals(self):
        """Parses stored interlace data, returns list of used interlace values.
        """
        out = [False]
        if self.Interlace:
            out.append(True)
        return out

    def get_vals_str(self, str):
        """Returns all possible values of setting requested by input string.
        """
        if str == "level":
            return self.get_level_vals()
        if str == "filter":
            return self.get_filter_vals()
        if str == "mode":
            return self.get_mode_vals()
        if str == "bitdepth":
            return self.get_bitdepth_vals()
        if str == "interlace":
            return self.get_interlace_vals()
        else:
            warning("unknown value passed to PNGProcessor.get_vals_str()")

    def get_max_files(self, proc = 0):
        """Returns maximum number of files generated in optimization.

        Used for % feedback.
        """
        out = len(self.get_level_vals()) * len(self.get_filter_vals()) *\
              len(self.get_interlace_vals())
        # multiply by number of truecolor/gray/paletted imgs
        outmult = 0

        if self.TrueColor:
            outmult += 1

        if self.MinPalDepth == 1:
            outmult += 4
        elif self.MinPalDepth == 2:
            outmult += 3
        elif self.MinPalDepth == 4:
            outmult += 2
        elif self.MinPalDepth == 8:
            outmult += 1

        if self.MinGrayDepth == 2:
            outmult += 3
        elif self.MinGrayDepth == 4:
            outmult += 2
        elif self.MinGrayDepth == 8:
            outmult += 1
        base = out * outmult
        if globals.ADVPNG:
            base += 2
        return base - self.ProcList[proc].OptMaxFilesDec 

    def reset(self, proc = 0):
        """Resets variables that are set separately for each file .
        """
        self.ProcList[proc].OptUsePal[1] = None
        self.ProcList[proc].OptUsePal[2] = None
        self.ProcList[proc].OptUsePal[4] = None
        self.ProcList[proc].OptUsePal[8] = None
        self.ProcList[proc].OptUseGray[2] = None
        self.ProcList[proc].OptUseGray[4] = None
        self.ProcList[proc].OptUseGray[8] = None
        self.ProcList[proc].OptMaxFilesDec = 0
        if globals.ADVPNG:
            self.ProcList[proc].DoAdvpng = True

    def generate_image(self, fnoext, reffile, lv, f, m, d, i, p, advpng=False):
        """Generates image with given settings.
        """
        outfile = (globals.TEMPDIR + fnoext + str(i) + m + str(lv) + f +
                   str(d) + str(advpng) + ".png")
        set = PNGSettings(lv, f, m, d, i, advpng)
        lossless = False
        if m == "truecolor":
            lossless = True
        self.ProcList[p].FCount += 1.0
        percent = self.ProcList[p].FCount / self.get_max_files(p) * 100
        output("png " + str(p) + " " + str(percent) +" %",3)
        return TempImage(outfile, "png", set, reffile, lossless)

    def helper(self, level, filter, ilace, reffile, fnoext, proc = 0):
        """Inner part of PNG generation.

        Taken out of inner loop of process_png.
        This generates PNG files using specified settings in all allowed
        modes, e.g. truecolor, paletted, grayscale, bitmap.
        """
        outfilelist = []
        #truecolor image
        if self.TrueColor:
            #advpng image
            if self.ProcList[proc].DoAdvpng:
                outfilelist.append(self.generate_image(fnoext, reffile, level,
                                   filter, "truecolor", 24, False, proc, True))
            #normal image
            outfilelist.append(self.generate_image(fnoext, reffile, level,
                               filter, "truecolor", 24, ilace, proc))
        #paletted image/s
        if self.MinPalDepth:
            depth = self.MinPalDepth
            while depth <= 8:
                if ((self.Optimizations >= 1) and
                    (self.ProcList[proc].OptUsePal[depth] == False)):
                    #OPTIMIZATION: See a few lines below.
                    self.ProcList[proc].OptMaxFilesDec += 1
                    depth *= 2
                    continue
                #advpng image
                if self.ProcList[proc].DoAdvpng:
                    outfilelist.append(self.generate_image(fnoext, reffile, 
                                       level, filter, "palette", depth, False,
                                       proc, True))
                #normal image
                outfilelist.append(self.generate_image(fnoext, reffile, level,
                                   filter, "palette", depth, ilace, proc))
                #OPTIMIZATION: As all paletted PNG images with same depth
                #store identical image data, in quality mode we need to check
                #if the image has at least given minquality only once
                #and we don't need to generate paletted images of that depth
                #again.
                #TODO: Same optimization for GIF, also no need to do
                #                                 interlaced/optimized JPG
                #                                 if quality is not enough
                if (self.Optimizations >= 1 and globals.TESTER.Mode == "q"
                    and self.ProcList[proc].OptUsePal[depth] == None):
                    if outfilelist[-1].Quality<globals.TESTER.abs_min_quality():
                        self.ProcList[proc].OptUsePal[depth] = False
                    else:
                        self.ProcList[proc].OptUsePal[depth] = True
                depth *= 2
        if self.ProcList[proc].DoAdvpng:
            self.ProcList[proc].DoAdvpng = False
        #grayscale image/s
        if self.MinGrayDepth:
            depthlist=[2, 4, 8]
            #find the index of given mindepth in list and start from there
            for d in depthlist[depthlist.index(self.MinGrayDepth):]:
                if ((self.Optimizations >= 1) and 
                    (self.ProcList[proc].OptUseGray[d] == False)):
                    #OPTIMIZATION: See a few lines below.
                    self.ProcList[proc].OptMaxFilesDec += 1
                    continue
                outfilelist.append(self.generate_image(fnoext, reffile, level,
                                   filter, "gray", d, ilace, proc))
                #OPTIMIZATION: As all grayscale PNG images with same depth
                #store identical image data, in quality mode we need to check
                #if the image has at least given minquality only once
                #and we don't need to generate grayscale images of that depth
                #again.
                #TODO:Same optimization for grayscale JPG
                if (self.Optimizations >= 1 and globals.TESTER.Mode == "q"
                    and self.ProcList[proc].OptUseGray[d] == None):
                    if outfilelist[-1].Quality<globals.TESTER.abs_min_quality():
                        self.ProcList[proc].OptUseGray[d] = False
                    else:
                        self.ProcList[proc].OptUseGray[d] = True
        return outfilelist                   
      
    def process(self, reffile, fnoext, queue=None, proc = 0):
        """Generates temporary png files and returns a list of them.
        """
        self.reset(proc)
        outfilelist = []
        output(str(proc) + " processing png")
        self.ProcList[proc].FCount = 0.0
        for level in self.get_level_vals(): 
            for f in self.get_filter_vals(): 
                for i in self.get_interlace_vals():
                    outfilelist.extend(self.helper(level, f, i, reffile,
                                       fnoext, proc))
        if queue is not None:
            queue.put(outfilelist)
        else:
            return outfilelist
