# Copyright (C) 2009 Ferdinand Majerech
# This file is part of IMGCrush
# For conditions of distribution and use, see copyright notice in LICENSE.txt

import getopt
import subprocess
from math import *
import sys
import globals
import re
import os

def clamp(num, minval, maxval):
    """Clamps the given number by given range.

    If the number is smaller than minimum, the minimum is returned,
    it it's larger than maximum, the maximum is returned,
    otherwise, the unchanged number is returned.
    """
    if(num < minval):
        return minval
    elif(num > maxval):
        return maxval
    else:                                                         
        return num

def run_bash(cmd, err=False):
    """Executes Bash commands and returns their output.

    If err is true, output is returned from stderr instead of stdout.
    """
    p = subprocess.Popen(cmd, shell = True, stdout = subprocess.PIPE, 
                         stderr = subprocess.PIPE)
    out, stderr = p.communicate()
    if err:
        return stderr.strip()
    return out.strip()

def clean_up():
    """Cleans the temporary directory when the program finishes work.
    """
    subprocess.call("rm -r " + globals.TEMPDIR, shell = True)

def file_size(filename):
    """Returns size of given file in bytes.
    """
    try:
        return int(run_bash("stat -c%s '" + filename + "'"))
    except ValueError:
        error("Can't determine size of a temporary file, probably because "+
              "it does not exist. Possible reason is not enough space"+
              "to generate temporary files.") 
        clean_up()
        sys.exit(-1)

def file_quality(reference, filename):
    """Returns quality of given file compared to given reference file.
    """
    try:
        outstr = run_bash("gm compare -metric mae '"+ reference + "' '" +
                            filename + "'" )
        for line in outstr.split("\n"):
            if "Total" in line:
                #matches any positive number (float or integer)
                floatre = re.compile("\d\.?\d*")
                errorstr = floatre.findall(line)[0]
                return 100.0 - 100 * float(errorstr)
    except ValueError:
        error("Can't determine quality of a temporary file, probably "+ 
              "because it is corrupted. Possible reason is not enough "+
              "space to generate temporary files.")
        clean_up()
        sys.exit(-1)      

def file_convert(format, settings, filename, outfilename,
                sizeargs=[0, 0, False]):
    """Converts the given file according to given parameters.
    """
    cmd = "convert "
    width = sizeargs[0]
    height = sizeargs[1]
    stretch = sizeargs[2]
    if width or height:
        cmd = cmd + "-resize "
        if width:
            cmd = cmd + str(width)
        if height:
            cmd = cmd + "x"+str(height)
        if width and height and stretch:
            cmd = cmd + "!"
        if not globals.NOSTRIP:
            cmd = cmd + " +profile \"*\" "
    if format == "jpg":
        cmd = cmd + "-sampling-factor " + str(settings.SamplingFactor) + " "
        cmd = cmd + "-quality " + str(settings.Quality) + " "
        if settings.Grayscale:
            cmd += "-colorspace Gray -colors 256 "
        if settings.Interlace:
            cmd = cmd + "-interlace Line "
        if settings.Optimize:
            cmd = cmd + "-define jpeg:optimize-coding=true "
        else:
            cmd = cmd + "-define jpeg:optimize-coding=false "
    if format == "png":
        if settings.Interlace:
            cmd = cmd + "-interlace Line "
        if settings.Mode == "palette":
            cmd = cmd + "-colors " + str(2 ** settings.BitDepth) + " "
        if settings.Mode == "gray":
            cmd = cmd + "-colorspace Gray -colors "\
                      + str(2 ** settings.BitDepth) + " "
        cmd = cmd + "-quality " + str(settings.Level) + str(settings.Filter)\
                  + " "
    if format == "gif":
        if settings.Interlace:
            cmd = cmd + "-interlace Line "
        cmd = cmd + "-colors " + str(2 ** settings.BitDepth) + " "
    cmd = cmd + "'" + filename + "' '" + outfilename + "'"
    subprocess.call(cmd, shell = True)
    if format == "png" and settings.Advpng:
        subprocess.call("advpng -4 -z '" + outfilename + "'" , shell = True)

def valid_number(input, default, command="UNKNOWN", min=0, max=100, 
                 isfloat=False):
    """Validates input as int and returns int.

    If input is invalid, prints warning about given command passing
    wrong value and returns given default value.
    """
    try:
        if isfloat:
            return clamp(float(input), min, max)
        else:
            return clamp(int(input), min, max)
    except ValueError:
        typestr = "int"
        if isfloat:
            typestr = "float"
        warning("Wrong value passed by "+ command + " : using default." +
                " Min value: " + str(min) + " Max value: " + str(max) +
                " Type of value: " + typestr + " Given value: " + input)
        return default

def valid_number_str(input, default, command="UNKNOWN", min=0, max=100,
                     isfloat=False):
    """Validates input as int and returns string.

    If input is invalid, prints warning about given command passing
    wrong value and returns given default value.
    """
    return str(valid_number(input, default, command, min, max, isfloat))

def valid_number_str_suffix(input, default, sufx, command,  min, max, 
                            isfloat=False):
    """Validates input as int with suffix and returns string.

    If input is invalid, prints warning about given command passing
    wrong value and returns given default value.
    """
    return valid_number_str(input[ :-len(sufx)], default, command +
                            " with suffix " + sufx, min, max, isfloat) + sufx

def output(input, verbosity = 2):
    """Print output
    """
    if globals.VERBOSITY >= verbosity:
        print input
        sys.stdout.flush()

def warning(input):
    """Print a warning
    """
    output("WARNING: " + input, 1)

def error(input):
    """Print an error
    """
    output("ERROR: " + input, 0)

